/*
 *  arch-tag: Implementation of podcast parse
 *
 *  Copyright (C) 2005 Renato Araujo Oliveira Filho - INdT <renato.filho@indt.org.br>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <libxml/entities.h>
#include <libxml/SAX.h>
#include <libxml/parserInternals.h>
#include <libgnomevfs/gnome-vfs.h>
#include <string.h>
#define __USE_XOPEN
#include <time.h>

#include "rb-podcast-parse.h" 

#define BUFFER_SIZE 256

struct RBPoadcastLoadContext
{
    guint in_unknown_elt;
    xmlParserCtxtPtr xmlctx;
    GString *prop_value;
    RBPodcastChannel *channel_data;
    RBPodcastItem *item_data;
    
    enum {
        RB_PODCAST_PARSER_STATE_START,
        RB_PODCAST_PARSER_STATE_RSS,
        RB_PODCAST_PARSER_STATE_CHANNEL,
        RB_PODCAST_PARSER_STATE_CHANNEL_PROPERTY,
        RB_PODCAST_PARSER_STATE_IMG,
        RB_PODCAST_PARSER_STATE_IMG_PROPERTY,
        RB_PODCAST_PARSER_STATE_ITEM,
        RB_PODCAST_PARSER_STATE_ITEM_PROPERTY,
        RB_PODCAST_PARSER_STATE_END,
    } state;
};

static gboolean rb_validate_channel_propert (const char *name);
static gboolean rb_validate_item_propert (const char *name);
static char* rb_podcast_parse_date(const char* date_str);
static void rb_podcast_parser_start_element (struct RBPoadcastLoadContext* ctx, const char *name, const char **attrs);
static void rb_podcast_parser_end_element (struct RBPoadcastLoadContext* ctx, const char *name);
static void rb_podcast_parser_characters (struct RBPoadcastLoadContext* ctx, const char *data, guint len);
static void rb_set_channel_value(struct RBPoadcastLoadContext* ctx, const char* name, const char* value);
static void rb_set_item_value(struct RBPoadcastLoadContext* ctx, const char* name, const char* value);

static RBPodcastItem *
rb_podcast_initializa_item()
{
    RBPodcastItem *data = g_new0(RBPodcastItem, 1);
    data->title = NULL;
    data->url = NULL;
    data->pub_date = NULL;
    return data;
}

static void
rb_set_channel_value(struct RBPoadcastLoadContext* ctx, const char* name, const char* value)
{
   if (!value)
       return;

   if (!strcmp(name, "title")) {
        ctx->channel_data->title = strdup(value);
   } else if (!strcmp(name, "language")) { 
        ctx->channel_data->lang = strdup(value);
   } else if (!strcmp(name, "itunes:subtitle")) {
        ctx->channel_data->subtitle = strdup(value);
   } else if (!strcmp(name, "itunes:summary")) { 
        ctx->channel_data->summary = strdup(value);
   } else if (!strcmp(name, "description")) {
        ctx->channel_data->description = strdup(value);
   } else if (!strcmp(name, "generator") ||
              !strcmp(name, "itunes:author")) {
        ctx->channel_data->author = strdup(value);
   } else if (!strcmp(name, "webMaster")) {
        ctx->channel_data->contact = strdup(value);
   } else if (!strcmp(name, "pubDate")) {
        ctx->channel_data->pub_date = strdup(value);
   } else if (!strcmp(name, "copyright")) {
        ctx->channel_data->copyright = strdup(value);
   } else if (!strcmp(name, "img")) {
        ctx->channel_data->img = strdup(value);
   };
}


static void
rb_set_item_value(struct RBPoadcastLoadContext* ctx, const char* name, const char* value)
{
    if (!strcmp(name, "title")) {
       ctx->item_data->title = strdup(value);
   } else if (!strcmp(name, "url")) {
       ctx->item_data->url = strdup(value);
   } else if (!strcmp(name, "pubDate")) {
       ctx->item_data->pub_date = rb_podcast_parse_date(value);
   };
}


static void
rb_insert_item(struct RBPoadcastLoadContext* ctx)
{
     RBPodcastItem *data = ctx->item_data;

     if ((data->url) && (data->url) && (data->pub_date)) {
        ctx->channel_data->lst_itens = g_list_append(ctx->channel_data->lst_itens, (void *) ctx->item_data);
     }
}

static gboolean rb_validate_channel_propert (const char *name)
{
    if (!strcmp(name, "title") || 
        !strcmp(name, "language") ||
        !strcmp(name, "itunes:subtitle") ||
        !strcmp(name, "itunes:summary") ||
        !strcmp(name, "description") || 
        !strcmp(name, "generator") ||
        !strcmp(name, "itunes:author") ||
        !strcmp(name, "webMaster") ||
        !strcmp(name, "lastBuildDate") ||
        !strcmp(name, "pubDate") ||
        !strcmp(name, "copyright"))
        return TRUE;
    else
        return FALSE;
    
}

static gboolean rb_validate_item_propert (const char *name)
{
    if (!strcmp(name, "title") ||
        !strcmp(name, "url") ||
        !strcmp(name, "pubDate"))
        return TRUE;
    else
        return FALSE;
}


static void 
rb_podcast_parser_start_element (struct RBPoadcastLoadContext* ctx, const char *name, const char **attrs)
{

    switch (ctx->state)
    {
        case RB_PODCAST_PARSER_STATE_START:
        {
            if (!strcmp(name, "rss")) {
                ctx->state = RB_PODCAST_PARSER_STATE_RSS;
            }
            else
                ctx->in_unknown_elt++;
            break;
        }
        
        case RB_PODCAST_PARSER_STATE_RSS:
        {
            if (!strcmp(name, "channel")) {
                ctx->state = RB_PODCAST_PARSER_STATE_CHANNEL;
            }
            else
                ctx->in_unknown_elt++;
            break;
        }

        case RB_PODCAST_PARSER_STATE_CHANNEL:
        {
            
            if (!strcmp(name, "image")) {
                ctx->state = RB_PODCAST_PARSER_STATE_IMG;
                break;
            } else  if (!strcmp(name, "item")) {
                ctx->item_data = rb_podcast_initializa_item(); // g_new0(RBPodcastItem, 1);
                ctx->state = RB_PODCAST_PARSER_STATE_ITEM;
                break;
            } else  if (!rb_validate_channel_propert (name)) {
                ctx->in_unknown_elt++;
                break;
            } 

            ctx->state = RB_PODCAST_PARSER_STATE_CHANNEL_PROPERTY;
            break;
        }

        case RB_PODCAST_PARSER_STATE_ITEM:
        {
            if (!strcmp(name, "enclosure")) {
                for (; *attrs; attrs +=2) {
                    if (!strcmp (*attrs, "url")) {
                        const char *url_value = *(attrs+1);
                        rb_set_item_value(ctx, "url", url_value);
                    }
                }
            } else  if (!rb_validate_item_propert (name)) {
                ctx->in_unknown_elt++;
                break;
            }
            

            ctx->state = RB_PODCAST_PARSER_STATE_ITEM_PROPERTY;
            break;
        }

        case RB_PODCAST_PARSER_STATE_IMG:
        {
            if (strcmp(name, "url") != 0) {
                ctx->in_unknown_elt++;
                break;
            }

            ctx->state = RB_PODCAST_PARSER_STATE_IMG_PROPERTY;
            break;
        }

        case RB_PODCAST_PARSER_STATE_CHANNEL_PROPERTY:
        case RB_PODCAST_PARSER_STATE_ITEM_PROPERTY:
        case RB_PODCAST_PARSER_STATE_IMG_PROPERTY:
        case RB_PODCAST_PARSER_STATE_END:
        break;
    }
}


static void
rb_podcast_parser_end_element (struct RBPoadcastLoadContext* ctx, 
                                const char *name)
{
//    if (*ctx->die == TRUE) {
//        xmlStopParser (ctx->xmlctx);
//        return;
//    }

    if (ctx->in_unknown_elt > 0) {
        ctx->in_unknown_elt--;
        return;
    }
               
    switch (ctx->state)
    {
        case RB_PODCAST_PARSER_STATE_START:
            ctx->state = RB_PODCAST_PARSER_STATE_END;
            break;
            
        case RB_PODCAST_PARSER_STATE_RSS:
            ctx->state = RB_PODCAST_PARSER_STATE_START;
            break;

        case RB_PODCAST_PARSER_STATE_CHANNEL:
            ctx->state = RB_PODCAST_PARSER_STATE_RSS;
            break;
            
        case RB_PODCAST_PARSER_STATE_CHANNEL_PROPERTY:
        {
            rb_set_channel_value(ctx, name, ctx->prop_value->str);
            ctx->state = RB_PODCAST_PARSER_STATE_CHANNEL;
            g_string_truncate (ctx->prop_value, 0);
            break;
        }

        case RB_PODCAST_PARSER_STATE_ITEM:
        {
            rb_insert_item(ctx);
            ctx->state = RB_PODCAST_PARSER_STATE_CHANNEL;
            break;
        }

        case RB_PODCAST_PARSER_STATE_ITEM_PROPERTY:
        {
            rb_set_item_value(ctx, name, ctx->prop_value->str);
            ctx->state = RB_PODCAST_PARSER_STATE_ITEM;
            g_string_truncate (ctx->prop_value, 0);
            break;
        }

        case RB_PODCAST_PARSER_STATE_IMG_PROPERTY:
        {
            rb_set_channel_value(ctx, "img", ctx->prop_value->str);
            ctx->state = RB_PODCAST_PARSER_STATE_IMG;
            g_string_truncate (ctx->prop_value, 0);
            break;
        }

        case RB_PODCAST_PARSER_STATE_IMG:
            ctx->state = RB_PODCAST_PARSER_STATE_CHANNEL;
            break;

        case RB_PODCAST_PARSER_STATE_END:
            break;
    }
}
    

static void
rb_podcast_parser_characters (struct RBPoadcastLoadContext* ctx, const char *data,
                                     guint len)
{
    switch (ctx->state)
    {
        case RB_PODCAST_PARSER_STATE_CHANNEL_PROPERTY:
        case RB_PODCAST_PARSER_STATE_ITEM_PROPERTY:
        case RB_PODCAST_PARSER_STATE_IMG_PROPERTY:            
            g_string_append_len (ctx->prop_value, data, len);
            break;
        case RB_PODCAST_PARSER_STATE_START:
        case RB_PODCAST_PARSER_STATE_IMG:
        case RB_PODCAST_PARSER_STATE_RSS:
        case RB_PODCAST_PARSER_STATE_CHANNEL:
        case RB_PODCAST_PARSER_STATE_ITEM:
        case RB_PODCAST_PARSER_STATE_END:
            break;
    }
}


void
rb_podcast_parse_load_feed(RBPodcastChannel *data, const char *file_name) {

    xmlParserCtxtPtr ctxt;
    xmlSAXHandlerPtr sax_handler = g_new0 (xmlSAXHandler, 1);
    GnomeVFSResult result;
    GnomeVFSHandle *read_handle;
    GnomeVFSFileSize bytes_read;
    GString *memory_buffer;
    guint buffer[BUFFER_SIZE];
    
    struct RBPoadcastLoadContext *ctx = g_new0 (struct RBPoadcastLoadContext, 1);

    ctx->in_unknown_elt = 0;
    ctx->channel_data = data;

    if (!gnome_vfs_initialized ()) {
        goto end_function;
    }
        
    result = gnome_vfs_open (&read_handle, file_name, GNOME_VFS_OPEN_READ);
    if (result != GNOME_VFS_OK) {
        goto end_function;
    }

    memory_buffer = g_string_new("");
    result = gnome_vfs_read(read_handle, &buffer, BUFFER_SIZE, &bytes_read);
    
    while (result == GNOME_VFS_OK) {
        g_string_append_len (memory_buffer, (const char *) buffer, bytes_read);
        result = gnome_vfs_read(read_handle, &buffer, BUFFER_SIZE, &bytes_read);
    }

    gnome_vfs_close(read_handle);

        
    //initializing parse 
    sax_handler->startElement = (startElementSAXFunc) rb_podcast_parser_start_element;
    sax_handler->endElement = (endElementSAXFunc) rb_podcast_parser_end_element;
    sax_handler->characters = (charactersSAXFunc) rb_podcast_parser_characters;

    ctx->prop_value= g_string_sized_new(512);
                            
    ctxt = xmlCreateMemoryParserCtxt(memory_buffer->str, memory_buffer->len);
    ctx->xmlctx = ctxt;
    xmlFree (ctxt->sax);
    ctxt->userData = ctx;
    ctxt->sax = sax_handler;
    xmlParseDocument (ctxt);
    ctxt->sax = NULL;
    xmlFreeParserCtxt (ctxt);

    g_string_free(ctx->prop_value, TRUE);
    g_string_free(memory_buffer, TRUE);

end_function:
    g_free(sax_handler);    
    g_free(ctx);
}

static char*
rb_podcast_parse_date(const char* date_str)
{
    struct tm tm;
    char buff[255];
    strptime(date_str, "%a, %d %b %Y %T", &tm);
    strftime(buff, sizeof(buff), "%d/%m/%Y %H:%M", &tm);
    return strdup(buff);
}
